// SPDX-License-Identifier: GPL-2.0+
#include <linux/module.h>	// included for all kernel modules
#include <linux/kernel.h>	// included for KERN_INFO
#include <linux/init.h>	  // included for __init and __exit macros
#include <linux/slab.h>
#include <linux/usb.h>
#include <linux/completion.h>
#include <linux/kthread.h>
#include <linux/mutex.h>
#include <linux/firmware.h>
#include <linux/uaccess.h>
#include "../include/memx_ioctl.h"
#include "memx_cascade_usb.h"

extern struct usb_driver memx_ai_driver;

static int memx_feature_trigger(struct transport_cmd *pCmd, struct memx_data *data, unsigned long arg)
{
	uint32_t cfg_header[2] = {0};
	int ret = 0;

	cfg_header[0] = (pCmd->SQ.opCode == ADMIN_CMD_SET_FEATURE) ? FWCFG_ID_SET_FEATURE : FWCFG_ID_GET_FEATURE;
	cfg_header[1] = sizeof(struct transport_cmd);

	memcpy(data->fw_wbuffer, cfg_header, 8);

	usb_fill_bulk_urb(data->fw_txurb, data->udev, usb_sndbulkpipe(data->udev, MEMX_FW_OUT_EP),
		data->fw_wbuffer, 8, memx_complete, data);

	/* send the data out the bulk port */
	if (usb_submit_urb(data->fw_txurb, GFP_KERNEL) < 0) {
		pr_err("Can't submit cfg dfp");
		ret = -ENOMEM;
		return ret;
	}

	if (!wait_for_completion_timeout(&data->fw_comp, msecs_to_jiffies(MXCNST_TIMEOUT30S))) {
		pr_err("wait cfg dfp timeout\n");
		ret = -ENOMEM;
		return ret;
	}

	memcpy(data->fw_wbuffer, pCmd, sizeof(struct transport_cmd));

	/* set up our urb */
	usb_fill_bulk_urb(data->fw_txurb, data->udev, usb_sndbulkpipe(data->udev, MEMX_FW_OUT_EP),
			data->fw_wbuffer, sizeof(struct transport_cmd), memx_complete, data);

	/* send the data out the bulk port */
	if (usb_submit_urb(data->fw_txurb, GFP_KERNEL) < 0) {
		pr_err("Can't submit TX URB");
		ret = -ENODEV;
		return ret;
	}

	if (!wait_for_completion_timeout(&data->fw_comp, msecs_to_jiffies(MXCNST_TIMEOUT30S))) {
		pr_err("wait tx timeout\n");
		ret = -ENODEV;
		return ret;
	}

	usb_fill_bulk_urb(data->fw_rxurb, data->udev, usb_rcvbulkpipe(data->udev, MEMX_FW_IN_EP),
		data->fw_rbuffer, MAX_OPS_SIZE, memx_fwrxcomplete, data);

	/* get the data in the bulk port */
	if (usb_submit_urb(data->fw_rxurb, GFP_KERNEL) < 0) {
		pr_err("Can't submit data read");
		ret = -ENOMEM;
		return ret;
	}

	if (!wait_for_completion_timeout(&data->fwrx_comp, msecs_to_jiffies(MXCNST_TIMEOUT30S))) {
		pr_err("wait data read timeout\n");
		ret = -ENOMEM;
		return ret;
	}

	return 0;
}

static long memx_feature_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	struct memx_data *data = file->private_data;
	struct usb_interface *interface = data->interface;
	struct transport_cmd feature_cmd = {0};
	long ret = 0;

	if (!interface)
		return -ENODEV;

	mutex_lock(&data->cfglock);

	if (copy_from_user(&feature_cmd, (struct transport_cmd *)arg, sizeof(struct transport_cmd))) {
		pr_err("MEMX_GET_DEVICE_FEATURE, copy_from_user fail\n");
		ret = -ENOMEM;
	} else {
		switch (cmd) {
		case MEMX_GET_DEVICE_FEATURE:
			if (feature_cmd.SQ.subOpCode == FID_DEVICE_THROUGHPUT) {
				feature_cmd.CQ.data[0]  = tx_time_us;
				feature_cmd.CQ.data[1]  = tx_size / KBYTE;
				feature_cmd.CQ.data[2] = rx_time_us;
				feature_cmd.CQ.data[3] = rx_size / KBYTE;
				feature_cmd.CQ.data[4] = udrv_throughput_info.stream_write_us;
				feature_cmd.CQ.data[5] = udrv_throughput_info.stream_write_kb;
				feature_cmd.CQ.data[6] = udrv_throughput_info.stream_read_us;
				feature_cmd.CQ.data[7] = udrv_throughput_info.stream_read_kb;
				tx_time_us = 0;
				tx_size = 0;
				rx_time_us = 0;
				rx_size = 0;
				udrv_throughput_info.stream_write_us = 0;
				udrv_throughput_info.stream_write_kb = 0;
				udrv_throughput_info.stream_read_us = 0;
				udrv_throughput_info.stream_read_kb = 0;
			} else {
				if (memx_feature_trigger(&feature_cmd, data, arg)) {
					pr_err("USB MEMX_GET_DEVICE_FEATURE failed!\n");
					ret = -EIO;
				}

				if (feature_cmd.SQ.subOpCode == FID_DEVICE_INFO) {
					char version[8] = VERSION;

					memcpy(&((struct transport_cmd*)data->fw_rbuffer)->CQ.data[7], version, sizeof(unsigned int));
					memcpy(&((struct transport_cmd*)data->fw_rbuffer)->CQ.data[8], &version[sizeof(unsigned int)], sizeof(unsigned int));
				}
			}
			if (copy_to_user((void __user*)arg, data->fw_rbuffer, sizeof(struct transport_cmd))) {
				pr_err("MEMX_GET_DEVICE_FEATURE, copy_to_user fail\n");
				ret = -ENOMEM;
				return ret;
			}
			break;
		case MEMX_SET_DEVICE_FEATURE:
			if (memx_feature_trigger(&feature_cmd, data, arg)) {
				pr_err("USB MEMX_GET_DEVICE_FEATURE failed!\n");
				ret = -EIO;
				break;
			}

			if (copy_to_user((void __user*)arg, data->fw_rbuffer, sizeof(struct transport_cmd))) {
				pr_err("MEMX_GET_DEVICE_FEATURE, copy_to_user fail\n");
				ret = -ENOMEM;
				return ret;
			}
			break;
		default:
			ret = -ENOTTY;
		}
	}

	mutex_unlock(&data->cfglock);

	return ret;
}

static int memx_feature_open(struct inode *inode, struct file *file)
{
	struct memx_data *dev;
	struct usb_interface *interface;

	/* get the interface from minor number and driver information */
	interface = usb_find_interface(&memx_ai_driver, iminor(inode));
	if (!interface)
		return -ENODEV;

	dev = usb_get_intfdata(interface);
	if (!dev)
		return -ENODEV;


	file->private_data = dev;

	return 0;
}

static s32 memx_feature_release(struct inode *inode, struct file *file)
{
#ifdef DEBUG
	u32 major = imajor(inode);
	u32 minor = iminor(inode);
#endif

#ifdef DEBUG
	pr_info(" feature_close: (%d-%d) success.\n", major, minor);
#endif
	return 0;
}

struct file_operations memx_feature_fops = {
owner: THIS_MODULE,
unlocked_ioctl : memx_feature_ioctl,
open : memx_feature_open,
release : memx_feature_release,
};
