// SPDX-License-Identifier: GPL-2.0+
#include <linux/fs.h>
#include "memx_fs.h"
#include "memx_fs_proc.h"
#include "memx_fs_sys.h"
#include "memx_cascade_debugfs.h"
#include "memx_fw_log.h"

#define MEMX_NUM_OF_FS_CMDS (sizeof(g_fs_cmd_tbl) / sizeof(struct memx_fs_cmd_cb))

static struct memx_fs_cmd_cb g_fs_cmd_tbl[] = {
	{ "fwlog", FS_CMD_FWLOG_ARGC, memx_fs_cmd_handler},
	{ "read",  FS_CMD_READ_ARGC,  memx_fs_cmd_handler},
	{ "write", FS_CMD_WRITE_ARGC, memx_fs_cmd_handler},
	{ "memx0", FS_CMD_MEMX0_ARGC, memx_fs_cmd_handler},
	{ "memx1", FS_CMD_MEMX1_ARGC, memx_fs_cmd_handler},
	{ "memx2", FS_CMD_MEMX2_ARGC, memx_fs_cmd_handler},
	{ "memx3", FS_CMD_MEMX3_ARGC, memx_fs_cmd_handler},
	{ "memx6", FS_CMD_MEMX6_ARGC, memx_fs_cmd_handler},
	{ "memx7", FS_CMD_MEMX7_ARGC, memx_fs_cmd_handler},
	{ "memx8", FS_CMD_MEMX8_ARGC, memx_fs_cmd_handler},
	{ "rmtcmd", FS_CMD_RMTCMD_ARGC, memx_fs_cmd_handler},
};

s32 memx_fs_cmd_handler(struct memx_data *memx_dev, u8 argc, char **argv)
{
	u8 chip_id  = 0;
	u32 reg_addr = 0, reg_value = 0;
	//bool is_access_mpu = false;

	if (!memx_dev) {
		pr_err("fs_cmd_handler: memx_dev is NULL\n");
		return -EINVAL;
	}
	if (argc < FS_CMD_FWLOG_ARGC) {
		pr_err("fs_cmd_handler: invalid argc(%u), should be 2-4\n", argc);
		return -EINVAL;
	}
	if (!argv) {
		pr_err("fs_cmd_handler: argv is NULL\n");
		return -EINVAL;
	}

	if (argc >= FS_CMD_FWLOG_ARGC &&
		kstrtou8(argv[FS_CMD_FWLOG_ARGC - 1], 0, &chip_id) &&
		chip_id >= memx_dev->chipcnt) {
		pr_err("fs_cmd_handler: invalid chip id(%u), should be 0 - %u\n", chip_id, memx_dev->chipcnt);
		return -EINVAL;
	}

	if (((argc == FS_CMD_READ_ARGC) || (argc == FS_CMD_WRITE_ARGC) || (argc == FS_CMD_RMTCMD_ARGC)) && kstrtou32(argv[FS_CMD_READ_ARGC - 1], 0, &reg_addr)) {
		pr_err("fs_cmd_handler: invalid reg_addr:%s\n", argv[2]);
		return -EINVAL;
	}

	if (((argc == FS_CMD_WRITE_ARGC) || (argc == FS_CMD_RMTCMD_ARGC)) && kstrtou32(argv[FS_CMD_WRITE_ARGC - 1], 0, &reg_value)) {
		pr_err("fs_cmd_handler: invalid reg_value:%s\n", argv[3]);
		return -EINVAL;
	}

	switch (argc) {
	case FS_CMD_FWLOG_ARGC: {
		s32 ret = memx_fw_log_dump(memx_dev, chip_id);
		//pr_err("FS_CMD_FWLOG_ARGC\n");
		if (ret) {
			pr_err("fs_cmd_handler: try to dump chip_id(%u)'s fw_log fail(%d)\n", chip_id, ret);
			return ret;
		}
		pr_info("fs_cmd_handler: dump of chip_id(%u)'s fw_log success\n", chip_id);
	} break;
	case FS_CMD_READ_ARGC: {
		//pr_err("FS_CMD_READ_ARGC\n");
		memx_send_memxcmd(memx_dev, chip_id, MXCNST_MEMXR_CMD, reg_addr, 0);
	} break;
	case FS_CMD_WRITE_ARGC: {
		//pr_err("FS_CMD_WRITE_ARGC\n");
		memx_send_memxcmd(memx_dev, chip_id, MXCNST_MEMXW_CMD, reg_addr, reg_value);
	} break;

	case FS_CMD_MEMX0_ARGC:
	case FS_CMD_MEMX1_ARGC:
	case FS_CMD_MEMX2_ARGC:
	case FS_CMD_MEMX3_ARGC:
	case FS_CMD_MEMX6_ARGC:
	case FS_CMD_MEMX7_ARGC:
	case FS_CMD_MEMX8_ARGC: {
		u32 cmdbase = MXCNST_MEMX0_CMD;

		cmdbase = cmdbase + ((argc>>4) - (FS_CMD_MEMX0_ARGC>>4));
		//pr_err("FS_CMD_MEMX_ARGC 0x%08X\n", cmdbase);
		memx_send_memxcmd(memx_dev, chip_id, cmdbase, 0, 0);
	} break;
	case FS_CMD_RMTCMD_ARGC: {
		//pr_err("FS_CMD_RMTCMD_ARGC 0x%08X\n", reg_addr);
		memx_send_memxcmd(memx_dev, chip_id, reg_addr, reg_value, 0);
	} break;

	default: {
		pr_err("fs_cmd_handler: expected argc[2-4] but (%u)\n", argc);
		return -EINVAL;
	}
	}
	return 0;
}

s32 memx_fs_parse_cmd_and_exec(struct memx_data *memx_dev, const char __user *user_input_buf, size_t user_input_buf_size)
{
	s32 ret = -EINVAL;
	char *found = NULL;
	const char delimiters[] = {' ', '\0'}; // space, \0
	char *argv[4] = {NULL, NULL, NULL, NULL}; // cmd_name, chip_id, addr, val
	u8 argc = 0;
	u8 cmd_idx = 0;
	char *input_parser_buffer_ptr = NULL;

	if (memx_dev->fs.type == MEMX_FS_HIF_SYS) {
		input_parser_buffer_ptr = kstrdup(user_input_buf, GFP_KERNEL);
		if (!input_parser_buffer_ptr) {
			//pr_err("memx_fs_parser: kstrdup fail!\n");
			return ret;
		}
	} else {
		input_parser_buffer_ptr = memdup_user_nul(user_input_buf, user_input_buf_size);
		if (IS_ERR(input_parser_buffer_ptr)) {
			pr_err("memx_fs_parser: memdup_user_nul fail!\n");
			return PTR_ERR(input_parser_buffer_ptr);
		}
	}

	// parse command string by deleimiters and get the final argc
	while ((found = strsep(&input_parser_buffer_ptr, delimiters)) != NULL)
		argv[argc++] = found;

	if (argc < FS_CMD_FWLOG_ARGC) {
		pr_err("memx_fs_parser: argc should be te range in 2-4 but (%u)\n", argc);
		return ret;
	}


	// check argc and name meet requirement
	for (cmd_idx = 0; cmd_idx < MEMX_NUM_OF_FS_CMDS; cmd_idx++) {
		if ((argc == (g_fs_cmd_tbl[cmd_idx].expected_argc & 0xF)) &&
			(strncmp(argv[0], g_fs_cmd_tbl[cmd_idx].name, strlen(g_fs_cmd_tbl[cmd_idx].name)) == 0)) {
			ret = g_fs_cmd_tbl[cmd_idx].exec(memx_dev, g_fs_cmd_tbl[cmd_idx].expected_argc, &argv[0]);
			break;
		}
	}

	kfree(input_parser_buffer_ptr);
	return ret;
}

s32 memx_fs_init(struct memx_data *memx_dev)
{
	s32 ret = -EINVAL;
	int minor = 0;

	if (!memx_dev) {
		pr_err("%s: memx_dev is NULL\n", __func__);
		return ret;
	}

	/* Decide minor number by file open */

	switch (memx_dev->fs.type) {
	case MEMX_FS_HIF_PROC: {
#ifndef ANDROID
		char name[128];
		struct file *fp;

		for (minor = 0; minor < 128; minor++) {
			sprintf(name, "/proc/memx%d/cmd", minor);
			fp = filp_open(name, O_RDONLY, 0);
			if (IS_ERR(fp)) {
				pr_info("register for %s\n", name);
				memx_dev->minor_index = minor;
				break;

			} else {
				//pr_err("file existed %p\n", fp);
				filp_close(fp, NULL);
			}
		}
#endif
		ret = memx_fs_proc_init(memx_dev);
	} break;

	case MEMX_FS_HIF_SYS: {
#ifndef ANDROID
		char name[128];
		struct file *fp;

		for (minor = 0; minor < 128; minor++) {
			sprintf(name, "/sys/memx%d/cmd", minor);
			fp = filp_open(name, O_RDONLY, 0);
			if (IS_ERR(fp)) {
				pr_info("register for %s\n", name);
				memx_dev->minor_index = minor;
				break;

			} else {
				//pr_err("file existed %p\n", fp);
				filp_close(fp, NULL);
			}
		}
#endif
		ret = memx_fs_sys_init(memx_dev);
	} break;

	default:
		pr_err("%s: non support file system type(%u)\n", __func__, memx_dev->fs.type);
		return ret;
	}

	return ret;
}

void memx_fs_deinit(struct memx_data *memx_dev)
{
	if (!memx_dev) {
		pr_err("%s: memx_dev is NULL\n", __func__);
		return;
	}
	switch (memx_dev->fs.type) {
	case MEMX_FS_HIF_PROC: {
		memx_fs_proc_deinit(memx_dev);
	} break;
	case MEMX_FS_HIF_SYS: {
		memx_fs_sys_deinit(memx_dev);
	} break;
	default:
		pr_err("%s: non support file system type(%u)\n", __func__, memx_dev->fs.type);
	}
}
